/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 04\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: pgoblin.c,v 1.26 2008/01/22 01:16:33 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"

static void
usage(int ex) {
    fprintf(stderr, "pgoblin " VERS "  " DATE "\n"
"Usage:\n"
"pgoblin [-aANOqvV ][-d DataBase ][-h Hostname ][-p Port ][-U User ][-f QueryFile][-Eflag[env]] args\n"
"    -f QueryFile\n"
"    -d DataBase name\n"
"    -h Hostname of backend server\n"
"    -p Port\n"
"____-U Username\n"
"    -c COPY out\n"
"    -C COPY in\n"
"    -D perform query without result\n"
"    -s SELECT\n"
"    -S SELECT BINARY emulation\n"
"____-e QueryArgument\n"
"    -q quiet\n"
"    -v verbose\n"
"    -N new style\n"
"    -O old style\n"
"");
    exit(ex);
}

static char *
rediropt(pgoblin_main *options, char *optarg, int redirenv) {
    char *cr;
    if  (redirenv > 0) {
        if  (!(cr = getenv(optarg))) {
            ifBLIN_QV1(options->flags) warnx("No ENV %s", optarg);
            exit(EX_USAGE);
        }
    }else{
        cr = optarg;
    }
    return(cr);
}

static void
cmdin(pgoblin_main *options, int redirenv, pgoblin_prog *pgm, pgoblin_cmd cmd, char *optarg) {
    if  (pgm->curr < pgm->maxx) {
        pgm->code[pgm->curr].cmd = cmd;
        pgm->code[pgm->curr].e.lit = rediropt(options, optarg, redirenv);
        pgm->curr++;
    } else {
        errx(EX_SOFTWARE, "Programm too big");
}   }

static mife_descriptor *
popin(pgoblin_main *options) {
    if  (options->stack[options->sp])
        return(mife_open(PGOBLIN_MIFEFLAGS, options->stack[options->sp++]));
    return(mife_opef(PGOBLIN_MIFEFLAGS, fileno(stdin)));
}

int
main(int argc, char *argv[], char **envp) {
    pgoblin_main *options;
    int c, redirenv, ex = EX_OK;
    u_int32_t tmp;
    char *o;
    mife_descriptor *in;

    if  (!(options = pgoblin_init(BLIN_VER1|PGOBLIN_OLDSTYLE, fileno(stdout), NULL, NULL))) {
        errx(EX_UNAVAILABLE, "Malloc failed #1\n");
    }
    if  (!(options->stack = malloc(sizeof(char**) * argc))) {
        errx(EX_UNAVAILABLE, "Malloc failed #2\n");
    }
    options->sp = 0;
    options->stack[0] = NULL;
    options->envp = envp;
    for ( redirenv = 0
        ; (c = getopt(argc, argv, "aAc:C:d:D:e:Ef:Fh:INOp:qs:S:t:U:vVw:?")) != EOF
        ; redirenv--
        ) {
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case 'a': options->flags |= PGOBLIN_COPEND_PASS;                                         break;
        case 'A': options->flags |= PGOBLIN_COPEND_APND;                                         break;
        case 'B':                                                                                break;
        case 'c': cmdin(options, redirenv, options->pgm, PGOBLIN_copyout, optarg);               break;
        case 'C': cmdin(options, redirenv, options->pgm, PGOBLIN_copyin, optarg);                break;
        case 'd': options->conn[0].dbname = rediropt(options, optarg, redirenv);                 break;
        case 'D': cmdin(options, redirenv, options->pgm, PGOBLIN_perform, optarg);               break;
        case 'e': in = mife_opem( PGOBLIN_MIFEFLAGS | MIFE_EOFL
                                , rediropt(options, optarg, redirenv)
                                , 0
                                );
                ; if  (!in) err(EX_DATAERR, "mife_opem -e");
                ; ex = pgoblin_parser(options, in);
                ; mife_close(in);
                ; if  (ex) err(EX_DATAERR, "Syntax?");
                ; break;
        case 'E': redirenv = 2;                                                                  break;
        case 'f': o = rediropt(options, optarg, redirenv);
                ; if  (o) {
                ;     options->stack[options->sp++] = o;
                ;     options->stack[options->sp] = NULL;
                ; }
                ; break;
        case 'F': if  (options->conn[0].flags & PGOBLIN_INT64TIME) {
                ;     options->conn[0].flags |= PGOBLIN_NEEDINTIME;
                ;     options->conn[0].flags &= ~PGOBLIN_INT64TIME;
                ; } else {
                ;     options->conn[0].flags |= PGOBLIN_FLOATTIME;
                ; }
                ; break;
        case 'h': options->conn[0].host = rediropt(options, optarg, redirenv);                   break;
        case 'I': if  (options->conn[0].flags & PGOBLIN_FLOATTIME) {
                ;     options->conn[0].flags |= PGOBLIN_NEEDINTIME;
                ;     options->conn[0].flags &= ~PGOBLIN_FLOATTIME;
                ; } else {
                ;     options->conn[0].flags |= PGOBLIN_INT64TIME;
                ; }
                ; break;
        case 'N': options->flags &= ~PGOBLIN_OLDSTYLE;                                           break;
        case 'O': options->flags |= PGOBLIN_OLDSTYLE;                                            break;
        case 'p': options->conn[0].port = rediropt(options, optarg, redirenv);                   break;
        case 'q': options->flags &= ~BLIN_VERMASK;                                               break;
        case 's': cmdin(options, redirenv, options->pgm, PGOBLIN_strselect, optarg);             break;
        case 'S': cmdin(options, redirenv, options->pgm, PGOBLIN_binselect, optarg);             break;
        case 't': options->maxtmout = strtoul(rediropt(options, optarg, redirenv), NULL, 0) * 1000;
                ; break;
        case 'U': options->conn[0].username = rediropt(options, optarg, redirenv);               break;
        case 'v': BLIN_VERBOSE(options->flags);                                                  break;
        case 'V': options->flags |= PGOBLIN_VERIFY;                                              break;
        case 'w': tmp = options->maxtmout = strtoul(rediropt(options, optarg, redirenv), NULL, 0);
                ; options->flags |= ((tmp > 255) ? 255 : tmp);
                ; break;
        case '?': usage(EX_OK);
        default:  usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (options->flags & PGOBLIN_VERIFY) pgoblin_verify(options);
    options->argv = argv;
    if  ((argc > 0) && argv && *argv && **argv) {
        options->stack[options->sp++] = *(argv++);
        options->stack[options->sp] = NULL;
    }
    options->sp = 0;
    if  (!(in = popin(options))) {
        ifBLIN_QV1(options->flags) warn("Programm");
        ex = EX_IOERR;
    } else if (!options->pgm->curr && pgoblin_parser(options, in)) {
        ifBLIN_QV1(options->flags) warnx("Parser failed");
        ex = EX_DATAERR;
    } else {
        if  ((options->io[0].mife = popin(options))) options->io[0].flags |= PGOBLIN_MIFEDES;
        if  (options->flags & PGOBLIN_VERIFY) pgoblin_Vdump(options, options->pgm);
        if  ((ex = pgoblin_execute(options))) ifBLIN_QV1(options->flags) warnx("Execute failed");
    }
    exit(ex);
}
