/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: pgoblin.h,v 1.32 2008/01/22 01:16:33 babolo Exp $
 */

#define PGOBLIN_REGSIZE   64
#define PGOBLIN_PROGSIZE  8191
#define PGOBLIN_MIFEFLAGS MIFE_FULL

/* For pgoblin_io.flags */
#define  PGOBLIN_EMPTY     0x000000
/*            For         pgoblin_syntax.ch[]  pgoblin_syntax.a[].flags          */
/*                                         \    /                                */
/*               pgoblin_io.flags ---\      |  |      /--- pgoblin_main.exeflags */
/*                                     \    |  |    /                            */
#define  PGOBLIN_PATHPARM  0x000001  /*  +  +     :  */
#define  PGOBLIN_ISIN      0x000001  /*        +     */
#define  PGOBLIN_TEXTPARM  0x000002  /*  +  +     :  */
#define  PGOBLIN_MIFEDES   0x000004  /*  +  +     -  */
#define  PGOBLIN_CTL       0x000004  /*        +     */
#define  PGOBLIN_PQRESULT  0x000008  /*  +  +     :  */
#define  PGOBLIN_OUTSET    0x000010  /*  +  +  +  :  */
#define  PGOBLIN_BINPARM   0x000020  /*  +           */
#define  PGOBLIN_CONNECT   0x000020  /*     +  +     */
#define  PGOBLIN_FORKOUT   0x000020  /*           +  */
#define  PGOBLIN_JOB       0x000040  /*     +  +     */
#define  PGOBLIN_FORKIN    0x000040  /*           +  */
#define  PGOBLIN_STYLE     0x000080  /*     +  +     */
#define  PGOBLIN_FORKCTL   0x000080  /*           +  */
#define  PGOBLIN_ANYIVL    (PGOBLIN_PATHPARM | PGOBLIN_TEXTPARM | PGOBLIN_MIFEDES | PGOBLIN_PQRESULT)
/**** First byte ends ****/
#define  PGOBLIN_EXTENDED  0x000100  /*     -  +     */
#define  PGOBLIN_STYLEN    0x003000  /*     -  +     */
#define  PGOBLIN_FORKED    0x000800  /*     -     #  */
#define  PGOBLIN_TRAPDEF   0x001000  /*     -     #  */
#define  PGOBLIN_INTRAP    0x002000  /*     -     #  */
#define  PGOBLIN_END       0x004000  /*     -     #  */
#define  PGOBLIN_ERROR     0x008000  /*     -     #  */
#define  PGOBLIN_EXTABLE   0x00FF00  /*     -     #  */
/**** Second byte ends ****/
#define  PGOBLIN_FREEPATH  0x010000  /*  +  -     +  */
#define  PGOBLIN_FREETEXT  0x020000  /*  +  -     +  */
#define  PGOBLIN_FREEMIFE  0x040000  /*  +  -        */
#define  PGOBLIN_FREEPQRE  0x080000  /*  +  -     +  */
#define  PGOBLIN_FREEOUTS  0x100000  /*  +  -        */
#define  PGOBLIN_LEXLVUP   0x100000  /*     -  +     */
#define  PGOBLIN_FILEOVER  0x200000  /*  +  -        */
#define  PGOBLIN_LEXLDOWN  0x200000  /*     -  +     */
#define  PGOBLIN_OPEN      0x200000  /*     -     +  */
#define  PGOBLIN_NOLITER   0x400000  /*     -  +     */
#define  PGOBLIN_IGNORE    0x800000  /*     -  +  +  */
#define  PGOBLIN_ILLEGAL   (PGOBLIN_IGNORE | PGOBLIN_NOLITER)
/**** Third byte ends ****/

/***************************/
/* pgoblin's command codes */
/***************************/


typedef enum pgoblin_command
{ PGOBLIN_extended   = 0x00
, PGOBLIN_null       = 0x01
, PGOBLIN_cont       = 0x02
, PGOBLIN_bang       = 0x03

, PGOBLIN_connect    = 0x10
, PGOBLIN_listen     = 0x11
, PGOBLIN_exec       = 0x12
, PGOBLIN_trap       = 0x13
, PGOBLIN_call       = 0x14
, PGOBLIN_include    = 0x15
, PGOBLIN_echo       = 0x18
, PGOBLIN_file       = 0x19
, PGOBLIN_cat        = 0x1A

, PGOBLIN_perform    = 0x20
, PGOBLIN_binpresel  = 0x21
, PGOBLIN_select     = 0x22
, PGOBLIN_binselect  = 0x23
, PGOBLIN_strselect  = 0x24
, PGOBLIN_copyin     = 0x28
, PGOBLIN_copyout    = 0x29
, PGOBLIN_getstr     = 0x2A

, PGOBLIN_begin   = PGOBLIN_EXTENDED | 0x00
, PGOBLIN_end     = PGOBLIN_EXTENDED | 0x01
, PGOBLIN_fork    = PGOBLIN_EXTENDED | 0x02
, PGOBLIN_wait    = PGOBLIN_EXTENDED | 0x03
, PGOBLIN_exit    = PGOBLIN_EXTENDED | 0x04
, PGOBLIN_return  = PGOBLIN_EXTENDED | 0x05
, PGOBLIN_display = PGOBLIN_EXTENDED | 0x06
, PGOBLIN_pipe    = PGOBLIN_EXTENDED | 0x07
, PGOBLIN_close   = PGOBLIN_EXTENDED | 0x08

, PGOBLIN_getarg  = PGOBLIN_EXTENDED | 0x10
, PGOBLIN_getenv  = PGOBLIN_EXTENDED | 0x11
, PGOBLIN_getget  = PGOBLIN_EXTENDED | 0x12
, PGOBLIN_getpost = PGOBLIN_EXTENDED | 0x13
, PGOBLIN_getproc = PGOBLIN_EXTENDED | 0x14
, PGOBLIN_getjob  = PGOBLIN_EXTENDED | 0x15
, PGOBLIN_getspq  = PGOBLIN_EXTENDED | 0x16
, PGOBLIN_max     = PGOBLIN_EXTENDED | 0x17
} pgoblin_command;

typedef u_int32_t pgoblin_cmd;
#define PGOBLIN_COMMAND 0x000000FF
#define PGOBLIN_ARG1    0x00003F00
#define PGOBLIN_ARG2    0x000FC000
#define PGOBLIN_ARG3    0x03F00000
#define PGOBLIN_ARG4    0xFC000000
#define PGOBLIN_ARGMASK 0x0000003F
#define PGOBLIN_BPARG   6           /* Bits per argument */
#define PGOBLIN_ARG1S   8           /* 1st arg shift     */
#define PGOBLIN_ARG2S   14          /* 2nd arg shift     */
#define PGOBLIN_ARG3S   20          /* 3rd arg shift     */
#define PGOBLIN_ARG4S   26          /* 4th arg shift     */

typedef struct pgoblin_tuple {
    pgoblin_cmd cmd;     /*                                  */
    union {
        char       *lit; /*                 */
        ssize_t     sts; /*              */
        pgoblin_cmd cmd; /*                       */
    }           e;
} pgoblin_tuple;

typedef struct pgoblin_prog {
#   define        PGOBLIN_STRING_ID_LEN 16
    u_char        id[PGOBLIN_STRING_ID_LEN]; /*  "#pGoblin"     ASCII */
    u_int32_t     flags;
    ssize_t       maxx;                      /*                     */
    ssize_t       curr;                      /*   pgoblin_t               */
    pgoblin_tuple code[1];                   /*                           */
} pgoblin_prog;

typedef struct pgoblin_io {
    u_int32_t        flags;
    void            *text;
    char            *path;
    size_t           length;
    size_t           offset;
    pid_t            pid;
    mife_descriptor *mife;
    PGresult        *pq;
    int            (*wri)(int onu, const void *buf, size_t size);
    int            (*clo)(int onu);
    int              onu;
} pgoblin_io;

typedef struct pgoblin_conn {
    u_int32_t        flags;
#   define           PGOBLIN_INT64TIME    0x000010 /* DB known to have int64 timestamp */
#   define           PGOBLIN_FLOATTIME    0x000020 /* DB known to have float timestamp */
#   define           PGOBLIN_NEEDINTIME   0x000040 /* Ask DB for timestamp format      */
#   define           PGOBLIN_CLOSECONN    0x000080 /* Close this connection on exit    */
    char            *dbname;   /* For libpq         */
    char            *host;     /* For libpq         */
    char            *port;     /* For libpq         */
    char            *username; /* For libpq         */
    PGconn          *conn;
    u_char           intran;   /* Transaction level */
} pgoblin_conn;

typedef struct pgoblin_main {
    u_int32_t        flags;
#   define           PGOBLIN_FORK_WAIT    0x0000FF /* Child waits after fork                  */
#   define           PGOBLIN_VERIFY       0x000100 /* Verify analizer and dump parsed program */
#   define           PGOBLIN_COPEND_PASS  0x000200 /* Pass through unconsumed end of file     */
#   define           PGOBLIN_COPEND_APND  0x000400 /* Append EOF marker to COPY out file      */
#   define           PGOBLIN_OLDSTYLE     0x800000 /* Redirect execed output to 0 reg_in      */
    u_char           vers[4];                      /* Version                                 */
    u_char           delim[4];                     /* Delimiter                               */
    u_int32_t        exeflags;
    pgoblin_prog    *pgm;                          /* Programm                                */
    ssize_t          ppoint;                       /* Programm counter                        */
    int              lexlevel;                     /* To skip nested blocks                   */
    ssize_t          ctrap;                        /*    trap               */
    ssize_t          rtrap;                        /*     trap          */
    size_t           sp;                           /* File names stack pointer                */
    char           **stack;                        /* File names stack                        */
    u_int32_t        maxtmout;                     /* Max timeout for LISTEN                  */
    pgoblin_io       io   [PGOBLIN_REGSIZE];       /* I/O register set                        */
    pgoblin_conn     conn [PGOBLIN_REGSIZE];       /* Connections register set                */
    mular_descriptor*job  [PGOBLIN_REGSIZE];       /* Child processes register set            */
    void            *style[PGOBLIN_REGSIZE];       /* Tables styles register set              */
    char           **argv;
    char           **envp;
    int              pip0[2];
    int              pip1[2];
    int              pip2[2];
} pgoblin_main;

typedef struct pgoblin_job {
    ssize_t          cmdn;
    pid_t            pid;
    int              status;
#   define           PGOBLIN_NOSTATUS (-1)
} pgoblin_job;

typedef struct pgoblin_r {
    union {
        pgoblin_io        *io;
        pgoblin_conn      *c;
        mular_descriptor **j;
        void             **s; /* style    */
    }         g;
    int       r;
    u_int32_t c;
} pgoblin_r;

typedef struct pgoblin_syntax {
#   define SYMREGS 6
#   define BINREGS 4
    ssize_t   symparam;
    ssize_t   binparam;
    pgoblin_command max;
    int8_t    binparams[BINREGS];
    struct {
        u_int8_t  sym;
        int8_t    bin;
        u_int8_t  mask;
        char     *name;
    }         p[SYMREGS];
    struct pgoblin_syntax_a {
        u_int32_t flags;
        u_int8_t  ch[BINREGS];
        int     (*car)(pgoblin_main *options, pgoblin_r *r);
        char     *str;
    }         a[PGOBLIN_max];
} pgoblin_syntax;

extern const pgoblin_syntax pgoblin;
extern const char pgoblin_regn[PGOBLIN_REGSIZE];

void        pgoblin_verify __P((pgoblin_main *options));
pgoblin_main *pgoblin_init __P((u_int32_t flags, int outfd, mife_descriptor *in, pgoblin_prog *pgm));
int         pgoblin_parser __P((pgoblin_main *options, mife_descriptor *in));
void         pgoblin_Vdump __P((pgoblin_main *options, pgoblin_prog *pgm));
int        pgoblin_execute __P((pgoblin_main *options));
